/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-channel.c: GUIChannel definition and associated functions
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <fcntl.h>
#include <crypt.h>
#include <time.h>

#include <gnome.h>
#include <glade/glade.h>
#include <gtkhtml/gtkhtml.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <libgnomeui/gnome-window-icon.h>

#include <gtkhtml/htmlimage.h>
#include <gtkhtml/htmlengine-search.h>

#include "gui.h"
#include "gui-channel.h"
#include "gui-transfer.h"
#include "gui-prefs.h"
#include "gui-install.h"
#include "mirror.h"
#include "subscription.h"
#include <gal/widgets/e-scroll-frame.h>
#include <gal/shortcut-bar/e-shortcut-bar.h>

#include "libredcarpet.h"
#include "rc-dep-or.h"

#include "gui-util.h"
#include "gui-uri.h"
#include "gui-html.h"

#include "subscription.h"


#undef BROKEN_OLDER

/* Local callbacks */

static void sub_button_clicked(GtkWidget *widget, GUIChannel *channel);
static void unsub_button_clicked(GtkWidget *widget, GUIChannel *channel);
static void notebook_page_switched(GtkNotebook *notebook, 
                                   GtkNotebookPage *page, int page_num, 
                                   GUIChannel *channel);

static void channel_refresh_clicked(GtkWidget *item, GUIChannel *channel);

GUIChannel *
rc_gui_channel_new (RCChannel *base_ch)
{
    GUIChannel *n;

    g_assert (base_ch);

    n = g_new0 (GUIChannel, 1);
    n->rcchannel = *base_ch;    /* ick */

    return n;
}

void
rc_gui_channel_clear_data (GUIChannel *channel)
{
    g_return_if_fail(channel);
    
    if (channel->html_data) {
        g_byte_array_free(channel->html_data, TRUE);
        channel->html_data = NULL;
    }

    /* So I think this is what you want -- itp */
    g_slist_free (channel->available_updates);
    g_slist_free (channel->installed_packages);
    g_slist_free (channel->uninstalled_packages);
    g_slist_free (channel->older_packages);

} /* rc_gui_channel_clear_data */

/* This function will fill in a GByteArray of the channel's HTML info as well
   as fill in the subchannels field of the RCChannel. */
int
rc_gui_channel_fetch_data(GUIChannel *channel)
{
    RCChannel *rcchannel = RCCHANNEL(channel);
    GByteArray *data;
    char *url;
    GUITransfer *t;
    gchar *info_url = NULL;
    int err;

    RC_ENTRY;

    /* Set the subscription flag */
    /* This little if thingy here is a bit of a hack. Because
       gui_channel_subscribed always pulls it from the subscription file, even
       if you "subscribe" to a channel as non-root, whenever this function
       gets called, it'll return to you the real value (non-root can't write
       to this file to actually change the subscription). So, what we're gonna
       do is see if the channel->subscribed flag has been toggled by the
       "subscription" and if it is, don't bother to re-read from disk. However,
       if it isn't get the real value from disk. The only time this won't work
       correctly is if non-root is subscribed to a channel, becomes root,
       changes the subscriptions.xml file and then the non-root user 
       refreshes. But you'd be stupid if you did that, and fuck you if you 
       did: just use the damned GUI */
    if (!channel->subscribed || !appinfo->not_root)
        channel->subscribed = rc_gui_subscribed(rcchannel);

    if (channel->subscribed) {
        info_url = rcchannel->subs_file;
    } else {
        info_url = rcchannel->unsubs_file;
    }

    /* Fetch the info HTML */
    if (info_url) {
        /* Do nothing; we'll defer fetching and rendering the html until the
           user actually clicks on this channel */
    } else {
        /* No info HTML file? Well, let's create an array... */
        char *message;

        channel->html_data = g_byte_array_new();
        message = "No channel info available";
        channel->html_data = g_byte_array_append(
            channel->html_data, message, strlen(message));
    }

    /* Fetch the subchannel/package/dependency info and fill it in */
    t = gui_transfer_new(GUI_TRANSFER_FLAGS_BLOCK, appinfo->normal_cache);
    if (appinfo->splash) {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(e_splash_get_progress_bar (appinfo->splash)));
    } else {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(GET_WIDGET("single_progress")));
    }

    url = rc_gui_get_mirrored_url(rcchannel->pkginfo_file, rcchannel);
    data = gui_transfer_begin(t, url);
    g_free(url);

    if (gui_transfer_get_error(t)) {
        gboolean retry;

        retry = gui_transfer_error_box(
            t,
            _("<p>If you select cancel, Red Carpet will skip this channel and "
              "it will be unavailable for the rest of this session (or until "
              "you refresh the channel list)."));

        if (retry) {
            if (data)
                g_byte_array_free(data, FALSE);
            return rc_gui_channel_fetch_data(channel);
        }
        else {
            /* This channel thing is inaccessible, so we just skip it. */
            return -1;
        }
    }
    gtk_object_unref(GTK_OBJECT(t));

    data = g_byte_array_append(data, "\0", 1);

    err = rc_channel_parse(rcchannel, data->data,
                           rcchannel->pkginfo_compressed ? data->len : 0);
    g_byte_array_free(data, TRUE);
    if (err) {
        gchar *s;
        gboolean retry;

        /* An error occurred. */
        s = g_strdup(
            _("<p>An error occured trying to parse channel information.</p>"
              "<p>This may be caused either by an old version of Red Carpet "
              "that is unable to read newer data files or by an error on the "
              "server. If you cancel, Red Carpet will skip this channel and "
              "it will be unavailable for the rest of this session (or until "
              "you refresh the channel list).</p>"));
        retry = rc_gui_error_box(s, RC_ERROR_RETRY);
        g_free(s);
        if (retry) {
            return rc_gui_channel_fetch_data(channel);;
        }
        else {
            /* This channel is inaccessible, so we just skip it. */
            return -1;
        }
    }

#ifdef GUI_USE_PACKAGESETS
    /* Fetch the packageset data */
    url = rc_gui_get_mirrored_url(rcchannel->pkgset_file, rcchannel);
    t = gui_transfer_new(GUI_TRANSFER_FLAGS_BLOCK, appinfo->normal_cache);
    gui_transfer_set_progress(t, GTK_PROGRESS(GET_WIDGET("single_progress")));
    data = gui_transfer_begin(t, url);
    g_free(url);
    gtk_object_unref(GTK_OBJECT(t));

    /* If the transfer succeeds, then we have sets. otherwise, they're null */
    if (data && data->data) {
        data = g_byte_array_append(data, "\0", 1);

        rcchannel->package_sets = rc_package_set_parse (data->data, data->len);
        g_byte_array_free (data, TRUE);
    } else {
        g_byte_array_free (data, FALSE);
    }
#endif

    RC_EXIT;
    return 0;
} /* rc_gui_channel_fetch_data */

void
rc_gui_channel_fetch_all_data(void)
{
    GSList *i;
    GtkProgress *progress = NULL, *single_progress = NULL;
    GtkLabel *progress_label = NULL;
    gint count = 0;

    RC_ENTRY;

    if (!appinfo->splash) {
        progress = GTK_PROGRESS(GET_WIDGET("total_progress"));
        single_progress = GTK_PROGRESS(GET_WIDGET("single_progress"));
        progress_label = GTK_LABEL(GET_WIDGET("progress_label"));
    }

    if (progress) {
        gtk_progress_configure(
            GTK_PROGRESS (progress), 0.0, 0.0,
            (gfloat) g_slist_length (appinfo->channel_list));
    }

    i = appinfo->channel_list;
    while (i) {
        GUIChannel *c = i->data;
        char *label;

        rc_debug (RC_DEBUG_LEVEL_INFO, "Fetching %s\n", RCCHANNEL(c)->name);
        label = g_strdup_printf(
            _("Downloading channel data (%s)"), RCCHANNEL(c)->name);
        if (appinfo->splash) {
            e_splash_set_progress_text (appinfo->splash, label);
        } else {
            gtk_label_set_text (progress_label, label);
            gtk_progress_set_value (single_progress, 0.0);
        }

        g_free(label);

        if (rc_gui_channel_fetch_data(c)) {
            char *err;
            gboolean retry;

            err = g_strdup_printf(
                _("<p>An error has occurred while downloading channel "
                  "data.</p>"
                  "<p>Red Carpet was unable to fetch information for the "
                  "\"%s\" channel from the server. Red Carpet can continue "
                  "without this information, but you will be unable to access "
                  "that channel during this session (unless you refresh)."
                  "</p>"),
                RCCHANNEL(c)->name);
            retry = rc_gui_error_box(err, RC_ERROR_RETRY);
            g_free(err);
            if (retry)
                continue;
            else {
                /* This failed for some reason; skip this channel. */
                /* this is semi-dangerous */
                rc_debug(
                    RC_DEBUG_LEVEL_WARNING, 
                    "Channel %d failed to download, skipping\n", 
                    RCCHANNEL(c)->id);
                i = i->next;
                appinfo->channel_list = g_slist_remove(
                    appinfo->channel_list, c);
                continue;
            }
        }

        if (progress) {
            gtk_progress_set_format_string (single_progress, "%p%%");
            gtk_progress_set_value (single_progress, 0.0);
            gtk_progress_set_value (progress, (gfloat) ++count);
            gtk_widget_draw (GTK_WIDGET (progress), NULL);       
        }
        
        i = i->next;
    }

    /* Process Debian 'or' dependencies possibly in these channels we 
     * just downloaded.
     */
    rc_dep_or_process_system_and_channels(
        appinfo->system_packages, appinfo->channel_list);

    /* Now go through and update the lists, once we've figured out
     * what we're subcsribed to and the like.
     */
    i = appinfo->channel_list;
    while (i) {
        GUIChannel *c = i->data;
        rc_gui_channel_update_package_lists(c);
        i = i->next;
    }

    RC_EXIT;
} /* rc_gui_fetch_channel_all_data */

/* FIXME: I moved this out of rc-channel.c, because it doesn't really
 * belong there, and it needs knowledge of the packman, which it can
 * get here.  Someday this belongs somewhere more general than here,
 * but not rc-channel.c ... */

static RCPackage *
rc_find_best_package (RCPackageDep *pdep, RCChannelSList *chs)
{
    RCPackage *ret = NULL;

    while (chs) {
        RCChannel *ch = (RCChannel *) chs->data;
        RCPackage *found;
        found = pkginfo_find_package_with_constraint (ch,
                                                      pdep->spec.name,
                                                      pdep,
                                                      FALSE);
#if 0
        if (!found) {
            found = g_hash_table_lookup (ch->dep_table,
                                         &pdep->spec);
        }
#endif
        if (found) {
            if (ret) {
                if (rc_packman_version_compare (appinfo->pman, &found->spec,
                                                &ret->spec) > 0) {
                    ret = found;
                }
            } else {
                ret = found;
            }
        }
        chs = chs->next;
    }

    return ret;
}

void
rc_gui_channel_update_package_lists(GUIChannel *channel)
{
    RCPackageSList *channel_packages = NULL;
    GHashTable *hash;
    GSList *i;
    RCChannelSList *sublist = NULL;
    RCChannelSList *unsublist = NULL;

    /* Get the list of packages in our preferred subchannel */
    hash = rc_hash_table_copy (RCCHANNEL (channel)->packages, g_str_hash, g_str_equal);

    /* Get the sub channels */
    rc_gui_split_channel_list (appinfo->channel_list, &sublist, &unsublist);

    /* Split up the package list into updates, installed, and
       available lists */

    channel->available_updates = NULL;
    channel->installed_packages = NULL;
    channel->uninstalled_packages = NULL;
    channel->older_packages = NULL;

    i = appinfo->system_packages;
    while (i) {
        RCPackage *package = i->data;
        RCPackage *p;
        gboolean used = FALSE;

        p = g_hash_table_lookup(hash, RC_PACKAGE_SPEC(package)->name);

        if (!p) {
            i = i->next;
            continue;
        }

        package->section = p->section;

        channel->installed_packages = g_slist_prepend(
            channel->installed_packages, package);

        if (!package->hold && rc_packman_version_compare (
                appinfo->pman, RC_PACKAGE_SPEC (package),
                RC_PACKAGE_SPEC (p)) < 0) {
            /* There is an update for this package in this channel.
             * We need to make sure that there isn't another update for this 
             * same package in another channel.
             */
            RCPackage *other_pkg;
            RCPackageDep *di;

            di = rc_package_dep_new_from_spec(
                RC_PACKAGE_SPEC(p), RC_RELATION_GREATER);
            other_pkg = rc_find_best_package(di, sublist);
            if (!other_pkg) {
                GUIChannelSList *channels = appinfo->channel_list;
                gboolean doit = TRUE;
                while (channels) {
                    GUIChannel *other_ch = (GUIChannel *) channels->data;
                    if (other_ch->subscribed && 
                        rc_package_spec_slist_find_name (
                            other_ch->available_updates,
                            di->spec.name)) {
                        doit = FALSE;
                    }
                    channels = channels->next;
                }
                if (doit)
                    channel->available_updates = g_slist_prepend(
                        channel->available_updates, p);
                used = TRUE;
            }
            rc_package_dep_free (di);
        }

#ifdef BROKEN_OLDER
        if (used)
            g_hash_table_remove(hash, RC_PACKAGE_SPEC(package)->name);
        else
            channel->older_packages = g_slist_prepend
                (channel->older_packages, package);
#else
        g_hash_table_remove(hash, RC_PACKAGE_SPEC(package)->name);
#endif
        i = i->next;
    }

    channel->uninstalled_packages = rc_package_hash_table_by_spec_to_list(
        hash);
    g_hash_table_destroy(hash);

    g_slist_free(channel_packages);
    
    g_slist_free (sublist);
    g_slist_free (unsublist);
} /* rc_gui_channel_update_package_lists */


static void
rc_gui_channel_connect_signals (GUIChannel *channel)
{
    GtkWidget *w;

    /* Connect the "switch-page" signal to the notebook */
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("notebook")), "switch-page",
        GTK_SIGNAL_FUNC(notebook_page_switched), channel);

    /* Connect the "link_clicked" signal to the HTML widgets */
    gtk_signal_connect(
        GTK_OBJECT(appinfo->update_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->installed_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->available_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->info_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);

    /* Connect clicked and selected signals to the subscription toolbar and
       menu items */
    w = GET_WIDGET("subscribe1");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(sub_button_clicked), 
        channel);
    
    w = GET_WIDGET("unsubscribe1");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(unsub_button_clicked),
        channel);

    /* Show and connect the signal to the channel refresh menu item and
       hide the summary refresh menu item */
    rc_gui_set_refresh_item(REFRESH_CHANNEL);
    w = GET_WIDGET("channel_refresh");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(channel_refresh_clicked), 
        channel);
} /* rc_gui_channel_connect_signals */

static void
rc_gui_channel_disconnect_signals(GUIChannel *channel)
{
    GtkWidget *w;

    /* not an error condition */
    if (!channel)
        return;

    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("notebook")),
        GTK_SIGNAL_FUNC(notebook_page_switched), channel);
    
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->update_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->installed_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->available_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->info_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);

    w = GET_WIDGET("subscribe1");
    g_assert(w);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(w), GTK_SIGNAL_FUNC(sub_button_clicked), channel);
    
    w = GET_WIDGET("unsubscribe1");
    g_assert(w);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(w), GTK_SIGNAL_FUNC(unsub_button_clicked), channel);

    w = GET_WIDGET("channel_refresh");
    g_assert(w);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(w), GTK_SIGNAL_FUNC(channel_refresh_clicked), channel);

} /* rc_gui_channel_disconnect_signals */

void
rc_gui_channel_show(GUIChannel *channel)
{
    GtkWidget *notebook;
    GtkWidget *w;

    RC_ENTRY;

    g_return_if_fail(channel);

    /* Get the notebook object */
    notebook = GET_WIDGET("notebook");
    g_assert(notebook);

    /* Hide the package HTML */
    gtk_widget_hide(GET_WIDGET("package_vbox"));

    /* First check to see if we are subscribed and sensitize the buttons */
    if (channel->subscribed) {
        rc_gui_html_sensitize_subscribe(FALSE);
        rc_gui_html_sensitize_unsubscribe(TRUE);
        gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
        if (!getenv ("RC_DEBUG_TABS"))
            gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
    }
    else {
        rc_gui_html_sensitize_subscribe(TRUE);
        rc_gui_html_sensitize_unsubscribe(FALSE);
        if (!getenv ("RC_DEBUG_TABS"))
            gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
        gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
    }

    rc_gui_set_page ("minor_notebook", PAGE_DETAILS);

    /* Get the search menu item and sensitize it */
    gtk_widget_set_sensitive(GET_WIDGET("search"), TRUE);
    
    /* Populate the preferences tab. */
    /* We have no subchannels, so always hide this page. */
    w = gtk_notebook_get_nth_page(
                                  GTK_NOTEBOOK(GET_WIDGET("notebook")), PAGE_PREFERENCES);
    g_assert(w);
    gtk_widget_hide(w);
    
    /* If we are subscribed, then we want to make sure we have all of the
       tabs. If we are unsubscribed, then we just want to show the info page */
    
    if (channel->subscribed) {
        rc_gui_set_page ("notebook", PAGE_UPDATES);
    }
    else {
        rc_gui_set_page ("notebook", PAGE_INFO);
    }

    while (gtk_events_pending ())
        gtk_main_iteration ();

    if (channel->subscribed) {
        rc_gui_render_updates_html(channel);
    }
    else {
        rc_gui_render_info_html(channel);
    }
} /* rc_gui_channel_show */

void
rc_gui_channel_subscribe(GUIChannel *channel)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "Subscribing to: %s\n", RCCHANNEL(channel)->name);

    if (!appinfo->not_root)
        rc_gui_subscribe(RCCHANNEL(channel));

    channel->subscribed = TRUE;
    
    rc_gui_set_action_html(channel, ACTION_UPDATE, 0);
    rc_gui_channel_refresh_updates ();
    rc_gui_channel_show(channel);
} /* rc_gui_channel_subscribe */

void
rc_gui_channel_unsubscribe(GUIChannel *channel)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "Unsubscribing to: %s\n", RCCHANNEL(channel)->name);

    if (!appinfo->not_root)
        rc_gui_unsubscribe(RCCHANNEL(channel));

    channel->subscribed = FALSE;

    rc_gui_set_action_html(channel, ACTION_INFO, 0);
    rc_gui_channel_refresh_updates ();
    rc_gui_channel_show(channel);
} /* rc_gui_channel_unsubscribe */

/* CB */
static void
channel_refresh_clicked(GtkWidget *item, GUIChannel *channel)
{
    GtkWidget *window;

    /* Hide the package HTML */
    gtk_widget_hide(GET_WIDGET("package_vbox"));

    gtk_widget_set_sensitive(GET_WIDGET("channel_refresh"), FALSE);
    gtk_widget_set_sensitive(GET_WIDGET("search"), FALSE);
    if (channel->subscribed)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), FALSE);
    else
        gtk_widget_set_sensitive(GET_WIDGET("subscribe1"), FALSE);

    rc_gui_channel_clear_data(channel);

    gtk_progress_configure(
        GTK_PROGRESS(GET_WIDGET("single_progress")), 0.0, 0.0, 1.0);
    gtk_progress_configure(
        GTK_PROGRESS(GET_WIDGET("total_progress")), 0.0, 0.0, 1.0);
    gtk_widget_hide(GET_WIDGET("total_progress"));

    gtk_widget_show(GET_WIDGET("startup_pause"));
    gtk_widget_hide(GET_WIDGET("startup_resume"));

    window = GET_WIDGET("startup_window");
    gtk_label_set_text(
        GTK_LABEL(GET_WIDGET("progress_label")),
        _("Downloading channel data..."));
    gtk_window_set_modal(GTK_WINDOW(window), TRUE);
    gtk_widget_show(window);

    rc_gui_channel_fetch_data(channel);
    rc_gui_fetch_icon(channel);
    rc_gui_channel_update_package_lists(channel);

    gtk_label_set_text(GTK_LABEL(GET_WIDGET("progress_label")), "Finished.");
    gtk_widget_hide(window);
    gtk_window_set_modal(GTK_WINDOW(window), FALSE);
    gtk_widget_show(GET_WIDGET("total_progress"));

    gtk_widget_set_sensitive(GET_WIDGET("channel_refresh"), TRUE);
    gtk_widget_set_sensitive(GET_WIDGET("search"), TRUE);
    if (channel->subscribed)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);
    else
        gtk_widget_set_sensitive(GET_WIDGET("subscribe1"), TRUE);    
 
    rc_gui_channel_update_package_lists(channel);
    rc_gui_channel_show(channel);
} /* channel_refresh_clicked */

static void
notebook_page_switched(GtkNotebook *notebook, GtkNotebookPage *page,
                       int page_num, GUIChannel *channel)
{
    ActionType types[] = {
        ACTION_UPDATE,
        ACTION_REMOVE,
        ACTION_INSTALL,
        ACTION_PREFERENCES,
        ACTION_INFO,
        ACTION_SEARCH
    };

    if (types[page_num] == ACTION_UPDATE)
        rc_gui_render_updates_html(channel);
    else
        rc_gui_set_action_html(channel, types[page_num], 0);
} /* notebook_page_switched */

static void
sub_button_clicked(GtkWidget *widget, GUIChannel *channel)
{
    /* Ugh. */
    rc_gui_handle_uri(appinfo->info_html, "updater:maybe_subscribe", channel);

    /* Old code... */
    /* rc_gui_channel_subscribe(channel); */
} /* sub_button_clicked */

static void
unsub_button_clicked(GtkWidget *widget, GUIChannel *channel)
{
    rc_gui_channel_unsubscribe(channel);
} /* unsub_button_clicked */

/*
 * rc_gui_channel_refresh_updates
 *
 * Go through each subscribed channel and referesh the available/other updates
 */
void rc_gui_channel_refresh_updates (void)
{
    GUIChannelSList *chiter;

    chiter = appinfo->channel_list;
    while (chiter) {
        GUIChannel *guic = (GUIChannel *) chiter->data;

        if (guic->subscribed)
            rc_gui_channel_update_package_lists (guic);

        chiter = chiter->next;
    }
}

/*
 * rc_gui_populate_option_menu
 *
 * populates option menu with each individual subchannel
 * and its name
 */

void
rc_gui_channel_select(GUIChannel *channel)
{
    rc_gui_channel_disconnect_signals(appinfo->current_channel);
    appinfo->current_channel = channel;
    if (channel) {
        rc_gui_channel_connect_signals(channel);
        rc_gui_channel_show(channel);
    }
} /* rc_gui_channel_select */
