/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * gui-install.c: Routines for installing and removing packages in the GUI
 *
 * Copyright (C) 2000-2002 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include <glib.h>

#include <gnome.h>
#include <glade/glade.h>

#include <gui.h>
#include <gui-prefs.h>
#include <mirror.h>

#include "gui-cache.h"
#include "gui-download.h"
#include "gui-html.h"
#include "gui-install.h"
#include "gui-log.h"
#include "gui-log-display.h"
#include "gui-priority.h"
#include "gui-transfer.h"
#include "gui-util.h"

#include "rc-debug-misc.h"

#include "rc-resolver.h"

#define REMOVE_WARNING_THRESHOLD 6

typedef enum {
    DEPENDENCY_CHECK_INSTALL,
    DEPENDENCY_CHECK_REMOVE,
    DEPENDENCY_CHECK_VERIFY
} DependencyCheckType;

static void
verify_cancel_cb (GtkWidget *widget, gboolean *data)
{
    if (data)
        *data = FALSE;

    gtk_main_quit ();
} /* verify_cancel_cb */

static void
verify_ok_cb (GtkWidget *widget, gboolean *data)
{
    *data = TRUE;

    gtk_main_quit ();
} /* verify_ok_cb */

static void
dependency_failure(GUIChannel *current_channel, RCResolver *resolver)
{
    gboolean sensitized;
    GtkWidget *verify_cancel;

    sensitized = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("select_all"));
    rc_gui_sensitize_select_items(SELECT_NEITHER);
    rc_gui_set_menubar_sensitive (FALSE);
    
    rc_gui_render_dependency_failure(current_channel, resolver);

    gtk_widget_set_sensitive (GET_WIDGET ("verify_ok"), FALSE);
    verify_cancel = GET_WIDGET ("verify_cancel");
    gtk_widget_set_sensitive (verify_cancel, TRUE);
    
    gtk_signal_connect (GTK_OBJECT (verify_cancel), "clicked",
                        GTK_SIGNAL_FUNC (verify_cancel_cb),
                        NULL);
    gtk_widget_show (verify_cancel);
    
    rc_gui_set_page ("main_notebook", PAGE_VERIFY);
    
    gtk_main ();

    rc_gui_set_menubar_sensitive (TRUE);
    if (sensitized)
        rc_gui_sensitize_select_items(SELECT_ITEM_ALL | SELECT_ITEM_NONE);
} /* dependency_failure */

static gboolean
dependency_confirm(RCPackageSList    *requested_pkgs,
                   gboolean           install,
                   GUIChannel        *current_channel,
                   RCChannelSList    *subscribed_channels,
                   RCChannelSList    *unsubscribed_channels,
                   RCPackageSList    *added_pkgs,
                   RCPackageSList    *removed_pkgs,
                   RCResolverContext *solution)
{
    GtkWidget *verify_cancel;
    GtkWidget *verify_ok;
    gboolean go_on;
    gboolean unsub;
    gboolean can_go_on = TRUE;
    gboolean sensitized;

    if (!requested_pkgs && !removed_pkgs && !added_pkgs)
        can_go_on = FALSE;

    sensitized = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("select_all"));

    rc_gui_sensitize_select_items(SELECT_NEITHER);
    go_on = rc_gui_render_dependencies(
        requested_pkgs, removed_pkgs,
        added_pkgs,
        solution,
        current_channel,
        !install);
    
    if (!go_on) {
        if (sensitized)
            rc_gui_sensitize_select_items(SELECT_ITEM_ALL | SELECT_ITEM_NONE);
        else
            rc_gui_sensitize_select_items(SELECT_NEITHER);
        return FALSE;
    }

    verify_cancel = GET_WIDGET("verify_cancel");
    verify_ok = GET_WIDGET("verify_ok");
    gtk_widget_show(verify_cancel);
    gtk_widget_show(verify_ok);

    gtk_widget_set_sensitive (verify_ok, can_go_on ? TRUE : FALSE);

    gtk_signal_connect(
        GTK_OBJECT(verify_cancel), "clicked",
        GTK_SIGNAL_FUNC(verify_cancel_cb), &go_on);
    gtk_signal_connect(
        GTK_OBJECT(verify_ok), "clicked",
        GTK_SIGNAL_FUNC(verify_ok_cb), &go_on);

    rc_gui_set_page("main_notebook", PAGE_VERIFY);

    unsub = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("unsubscribe1"));
    rc_gui_set_menubar_sensitive (FALSE);

    gtk_main ();

    rc_gui_set_menubar_sensitive (TRUE);

    if (unsub)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);

    gtk_signal_disconnect_by_func(
        GTK_OBJECT(verify_cancel),
        GTK_SIGNAL_FUNC(verify_cancel_cb), &go_on);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(verify_ok),
        GTK_SIGNAL_FUNC(verify_ok_cb), &go_on);

    if (go_on) {
        int r;

        r = g_slist_length(removed_pkgs);
        if (!install)
            r += g_slist_length(requested_pkgs);

        /* If we're going to remove more than REMOVE_WARNING_THRESHOLD
         * packages, pop up a dialog and make sure the user really wants to
         * go through with the transaction.
         */
        if (r > REMOVE_WARNING_THRESHOLD) {
            int button;
            GtkWidget *d = GET_WIDGET("remove_dialog");
            
            gnome_dialog_close_hides(GNOME_DIALOG(d), TRUE);
            button = gnome_dialog_run_and_close(GNOME_DIALOG(d));

            go_on = button == 0 ? TRUE : FALSE;
        }
    }

    if (sensitized)
        rc_gui_sensitize_select_items(SELECT_ITEM_ALL | SELECT_ITEM_NONE);
    else
        rc_gui_sensitize_select_items(SELECT_NEITHER);
    return go_on;
} /* dependency_confirm */

typedef struct {
    RCPackageSList *added_packages;
    RCPackageSList *removed_packages;
    RCPackageSList *requested_packages;
} RCDepInfo;

static void
prepend_pkg (RCPackage *pkg, RCPackageStatus status, gpointer user_data)
{
    RCDepInfo *dep_info = user_data;
    RCPackage *p = rc_package_spec_slist_find_name(
        dep_info->requested_packages, pkg->spec.name);

    if (status == RC_PACKAGE_STATUS_TO_BE_INSTALLED && !p) {
        /* The package is marked to be installed, but is not in the
         * requested list of packages.  This means that it's a package
         * being pulled in by a dependency.
         */
        dep_info->added_packages = g_slist_prepend(
            dep_info->added_packages, pkg);
    }
    else if (status == RC_PACKAGE_STATUS_TO_BE_UNINSTALLED &&
             pkg->installed && !p) {
        /* The package is marked to be uninstalled and is already installed,
         * and it isn't in the list of requested packages (which would mean
         * it's an upgrade), so it must be removed.
         */
        dep_info->removed_packages = g_slist_prepend(
            dep_info->removed_packages, pkg);
    }
}

static void
prepend_pkg_pair (RCPackage *pkg_to_add,
                  RCPackageStatus status_to_add,
                  RCPackage *pkg_to_remove,
                  RCPackageStatus status_to_remove, 
                  gpointer user_data)
{
    RCDepInfo *dep_info = user_data;
    RCPackage *p_add = rc_package_spec_slist_find_name(dep_info->requested_packages, 
                                                       pkg_to_add->spec.name);
#if 0
    RCPackage *p_remove = rc_package_spec_slist_find_name(dep_info->requested_packages, 
                                                          pkg_to_remove->spec.name);
#endif

    if (status_to_add == RC_PACKAGE_STATUS_TO_BE_INSTALLED && !p_add) {
        /* The package is marked to be installed, but is not in the
         * requested list of packages.  This means that it's a package
         * being pulled in by a dependency.
         */
        dep_info->added_packages = g_slist_prepend(
            dep_info->added_packages, pkg_to_add);
    }

    /* We don't need to show the removal part of the upgrade */
#if 0    
    if (status_to_remove == RC_PACKAGE_STATUS_TO_BE_UNINSTALLED &&
        pkg_to_remove->installed && !p_remove) {
        /* The package is marked to be uninstalled and is already installed,
         * and it isn't in the list of requested packages (which would mean
         * it's an upgrade), so it must be removed.
         */
        dep_info->removed_packages = g_slist_prepend(
            dep_info->removed_packages, pkg_to_remove);
    }
#endif
}

static gboolean
dependency_check(RCPackageSList       *requested_pkgs,
                 DependencyCheckType   type,
                 GUIChannel           *current_channel,
                 RCChannelSList       *subscribed_channels,
                 RCChannelSList       *unsubscribed_channels,
                 RCPackageSList      **added_pkgs,
                 RCPackageSList      **removed_pkgs)
{
    RCResolver *resolver;
    RCResolverContext *solution;
    GSList *iter;
    RCDepInfo dep_info;
    gboolean success = TRUE;

    resolver = rc_resolver_new ();

    if (type == DEPENDENCY_CHECK_INSTALL) {
        rc_resolver_add_packages_to_install_from_slist(
            resolver, requested_pkgs);
    }
    else if (type == DEPENDENCY_CHECK_REMOVE) {
        rc_resolver_add_packages_to_remove_from_slist(
            resolver, requested_pkgs);
    }

    if (current_channel != NULL) {
        rc_resolver_set_current_channel (resolver, 
                                         (RCChannel *) current_channel);
    }

    for (iter = subscribed_channels; iter != NULL; iter = iter->next) {
        rc_resolver_add_subscribed_channel (resolver,
                                            (RCChannel *) iter->data);
    }

    if (rc_packman_get_capabilities(appinfo->pman) &
        RC_PACKMAN_CAP_VIRTUAL_CONFLICTS)
        rc_resolver_allow_virtual_conflicts(resolver, TRUE);

    if (type == DEPENDENCY_CHECK_VERIFY)
        rc_resolver_verify_system (resolver);
    else
        rc_resolver_resolve_dependencies (resolver);
    
    solution = resolver->best_context;

    if (getenv("RC_SPEW")) {
        GSList *i;

        printf("#### Complete queues ####\n");
        for (i = resolver->complete_queues; i; i = i->next) {
            RCResolverQueue *q = i->data;

            if (q->context == resolver->best_context)
                printf ("** THIS IS BEST **\n");
            rc_resolver_context_spew_info(q->context);
            printf("#########################\n");
        }

        printf("#### Invalid queues ####\n");
        for (i = resolver->invalid_queues; i; i = i->next) {
            RCResolverQueue *q = i->data;

            rc_resolver_context_spew_info(q->context);
            printf("#########################\n");
        }
    }
        
    if (solution == NULL) {
        /* There is no valid solution */
        dependency_failure(current_channel, resolver);
        success = FALSE;
        goto finish;
    }

    dep_info.added_packages = NULL;
    dep_info.removed_packages = NULL;
    dep_info.requested_packages = requested_pkgs;

    rc_resolver_context_foreach_install (solution, prepend_pkg, &dep_info);
    rc_resolver_context_foreach_upgrade (solution, prepend_pkg_pair, &dep_info);
    rc_resolver_context_foreach_uninstall (solution, prepend_pkg, &dep_info);

    dep_info.added_packages = g_slist_reverse(dep_info.added_packages);
    dep_info.removed_packages = g_slist_reverse(dep_info.removed_packages);

    if (!dependency_confirm(
            requested_pkgs,
            type == DEPENDENCY_CHECK_REMOVE ? FALSE : TRUE,
            current_channel, subscribed_channels,
            unsubscribed_channels, dep_info.added_packages,
            dep_info.removed_packages, solution)) {
        success = FALSE;
    }

    if (added_pkgs)
        *added_pkgs = dep_info.added_packages;
    else
        g_slist_free(dep_info.added_packages);

    if (removed_pkgs)
        *removed_pkgs = dep_info.removed_packages;
    else
        g_slist_free(dep_info.removed_packages);

finish:
    rc_resolver_free (resolver);

    return success;
} /* dependency_check */

static void
transact_start_cb (RCPackman *packman, guint total)
{
    GtkWidget *w = gtk_object_get_data (GTK_OBJECT (appinfo->download_html),
                             "total_progress");
    gtk_progress_configure (GTK_PROGRESS (w), 0.0, 0.0, (gfloat) total);
    rc_gui_set_progress (PROGRESS_TRANSACT, total, 0,
                         _("Beginning transaction..."));
} /* transact_start_cb */

static void
transact_step_cb(RCPackman *p, gint pos, RCPackmanStep step, gchar *name)
{
    GtkWidget *w;
    GUIInstallType itype;
    char *text;
    char *word = NULL;

    rc_debug (RC_DEBUG_LEVEL_INFO, "Transacting %d\n", pos);
    w = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");
    gtk_progress_set_value(GTK_PROGRESS(w), (gfloat) pos);

    switch (step) {
    case RC_PACKMAN_STEP_UNKNOWN:
        word = "Transmogrifying";
        break;
    case RC_PACKMAN_STEP_CONFIGURE:
        word = _("Configuring");
        break;
    case RC_PACKMAN_STEP_INSTALL:
        word = _("Installing");
        break;
    case RC_PACKMAN_STEP_REMOVE:
        word = _("Removing");
        break;
    default:
        g_assert_not_reached();
        break;
    }

    text = g_strdup_printf ("%s %s", word, name ? name : _("packages"));

    rc_gui_set_progress(PROGRESS_TRANSACT, 0, pos, text);

    itype = rc_gui_get_install_type();
    gui_log_append(
        appinfo->log[itype == GUI_INSTALL ? LOG_TRANSACT : LOG_REMOVE], 
        text);

    g_free(text);
} /* transact_step_cb */
    
static void
transact_progress_cb(RCPackman *p, guint amount, guint total)
{
    gtk_progress_set_value(
        GTK_PROGRESS(GET_WIDGET("single_progress")),
        ((gfloat) amount) / ((gfloat) total) * 100.0);
} /* transaction_progress_cb */

static void
transact_done_cb(RCPackman *p)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "Transacting done\n");
} /* transaction_done_cb */

static void
update_total_progress(GUITransfer *t, gchar *buf, int size, int sizes[])
{
    GtkWidget *w;
    char *s1, *s2, *str;

    w = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");

    s1 = rc_gui_format_size(sizes[0] + t->trans_size + t->offset);
    s2 = rc_gui_format_size(sizes[1]);
    
    str = g_strdup_printf(_("%%p%%%% (%s of %s)"), s1, s2);
    g_free(s1);
    g_free(s2);
    gtk_progress_set_format_string(GTK_PROGRESS(w), str);
    g_free(str);
    gtk_progress_set_value(
        GTK_PROGRESS(w), (gfloat) (sizes[0] + t->trans_size + t->offset));
} /* update_total_progress */

static gboolean
download_files(GSList *urls, guint64 total_size)
{
    GtkWidget *single_progress, *total_progress;
    gboolean unsub;
    int total_count;
    int current_count = 0;
    guint64 current_size = 0;
    GSList *iter;

    /* Reset the progress thingys */
    single_progress = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "single_progress");
    total_progress = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");

    gtk_progress_configure(GTK_PROGRESS (single_progress), 0.0, 0.0, 100.0);
    gtk_progress_configure(
        GTK_PROGRESS (total_progress), 0.0, 0.0, (gfloat) total_size);
    gtk_progress_set_format_string(GTK_PROGRESS(total_progress), "%p%%");

    /* Desensitize the OK button during download */
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_set_sensitive (GET_WIDGET("ok_button"), FALSE);

    /* Show/hide the appropriate buttons. This is more paranoid than it needs
       to be, but better safe than sorry. */
    gtk_widget_show(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));
    gtk_widget_show(GET_WIDGET("cancel_button"));

    total_count = g_slist_length(urls);

    rc_gui_set_progress(PROGRESS_DOWNLOAD, total_count, 0, NULL);

    unsub = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("unsubscribe1"));
    rc_gui_set_menubar_sensitive (FALSE);

    iter = urls;
    while (iter) {
        gchar *url = (gchar *) iter->data;
        GUITransfer *t;
        int sizes[2];
        char *s1, *s2, *str;
        char *tmp;

        gtk_progress_set_value(GTK_PROGRESS(single_progress), 0.0);
        gtk_progress_set_format_string(GTK_PROGRESS(single_progress), "%p%%");
        tmp = g_strdup_printf(_("Downloading %s"), strrchr(url, '/') + 1);
        rc_gui_set_progress(
            PROGRESS_DOWNLOAD, total_count, current_count, tmp);
        g_free(tmp);

        t = gui_transfer_new(
            GUI_TRANSFER_FLAGS_BLOCK | 
            GUI_TRANSFER_FLAGS_FLUSH_MEMORY |
            GUI_TRANSFER_FLAGS_RESUME_PARTIAL,
            appinfo->package_cache);
        gui_transfer_set_progress_bar(t, GTK_PROGRESS(single_progress));
        sizes[0] = current_size;
        sizes[1] = total_size;
        gtk_signal_connect(
            GTK_OBJECT(t), "file_data",
            GTK_SIGNAL_FUNC(update_total_progress), sizes);

        if (t->aborted) {
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            goto download_cleanup;
        }
        
        gui_transfer_begin(t, url);

        if (t->aborted) {
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            goto download_cleanup;
        }

        if (gui_transfer_get_error(t)) {
            gboolean retry;
            
            retry = gui_transfer_error_box(t, NULL);
            
            gtk_object_unref(GTK_OBJECT(t));

            if (retry) {
                /* Don't iterate. Just do this one over again. */
                continue;
            }
            else {
                rc_gui_set_page("main_notebook", PAGE_CHANNELS);
                goto download_cleanup;
            }
        }

        if (t->cached) {
            /* Cache hit */
            str = g_strdup_printf(_("%s pulled from cache"), url);
        }
        else {
            s1 = rc_gui_format_size((int)
                ((float) t->file_size / g_timer_elapsed(t->timer, NULL)));
            str = g_strdup_printf(
                _("%s downloaded %d bytes in %d seconds (%s/s)"),
                strrchr(url, '/') + 1, t->file_size, 
                (int) g_timer_elapsed(t->timer, NULL), s1);
            g_free(s1);
        }
        gui_log_append(appinfo->log[LOG_DOWNLOAD], str);
        g_free(str);

        current_count++;
        current_size += t->file_size;

        /* This is mainly for cache hits */
        s1 = rc_gui_format_size(current_size);
        s2 = rc_gui_format_size(total_size);
        str = g_strdup_printf(_("%%p%%%% (%s of %s)"), s1, s2);
        g_free(s1);
        g_free(s2);
        gtk_progress_set_format_string(
            GTK_PROGRESS(total_progress), str);
        gtk_progress_set_value(
            GTK_PROGRESS(total_progress), (gfloat) current_size);
        g_free(str);
 
        gtk_object_unref(GTK_OBJECT(t));

        iter = iter->next;
    }

    rc_gui_set_progress(PROGRESS_DOWNLOAD, total_count, total_count, NULL);

    return TRUE;

download_cleanup:
    rc_gui_set_menubar_sensitive (TRUE);
    if (unsub)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);
    return FALSE;
} /* download_files */

static gboolean
verify_packages(RCPackageSList *packages)
{
    GtkWidget *single, *total;
    int total_count;
    int current_count = 0;
    GSList *i;

    gtk_widget_hide(GET_WIDGET("cancel_button"));
    gtk_widget_hide(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));

    total_count = g_slist_length(packages);

    single = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "single_progress");
    total = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");

    gtk_progress_configure(GTK_PROGRESS(single), 0.0, 0.0, 100.0);
    gtk_progress_configure(
        GTK_PROGRESS(total), 0.0, 0.0, (gfloat) total_count);

    gtk_progress_set_format_string(GTK_PROGRESS(single), "%p%%");
    gtk_progress_set_format_string(GTK_PROGRESS(total), _("%p%% (%v of %u)"));

    i = packages;
    while (i) {
        RCPackage *package = i->data;
        RCVerificationSList *vers;
        RCVerificationStatus worst_status = RC_VERIFICATION_STATUS_PASS;
        gboolean gpg = FALSE;
        char *gpg_info = NULL;
        GSList *v;
        char *tmp;

        tmp = g_strdup_printf(
            _("Verifying %s"), rc_gui_get_pretty_name(package->spec.name));
        rc_gui_set_progress(PROGRESS_VERIFY, total_count, current_count, tmp);
        g_free(tmp);
        while (gtk_events_pending())
            gtk_main_iteration();

        vers = rc_packman_verify(appinfo->pman, package);
        v = vers;
        while (v) {
            RCVerification *ver = v->data;

            if (worst_status > ver->status)
                worst_status = ver->status;

            if (ver->type == RC_VERIFICATION_TYPE_GPG) {
                gpg = TRUE;
                gpg_info = ver->info;
            }

            v = v->next;
        }

        if (worst_status == RC_VERIFICATION_STATUS_FAIL) {
            char *text;
            char *tmp_fn;

            text = g_strdup_printf(
                _("Cryptographic verification of %s (%s) has FAILED."),
                rc_gui_get_pretty_name(package->spec.name), 
                package->spec.name);
            gui_log_append(appinfo->log[LOG_VERIFY], text);
            g_free(text);
            
            text = g_strdup_printf(
                _("Red Carpet's cryptographic verification of %s (%s) has "
                  "<b><font color=\"#ff0000\">failed</font></b>. This means "
                  "that the security of this package has somehow been "
                  "compromised and Red Carpet will not install it. The "
                  "downloaded package may have been corrupted and downloading "
                  "the package again may fix the problem. If it doesn't, "
                  "you can "
                  "<a href=\"updater:send_bug_report?Verification failed\">"
                  "send a problem report</a> to Ximian for analysis."),
                rc_gui_get_pretty_name(package->spec.name), 
                package->spec.name);
            rc_gui_error_box(text, RC_ERROR_ABORT);
            g_free(text);

            /* Move the package and signature out of the way so they can
               download it again. */
            if (package->package_filename) {
                tmp_fn = g_strdup_printf("%s.bad", package->package_filename);
                rename(package->package_filename, tmp_fn);
                g_free(tmp_fn);
            }

            if (package->signature_filename) {
                tmp_fn = g_strdup_printf("%s.bad", package->signature_filename);
                rename(package->signature_filename, tmp_fn);
                g_free(tmp_fn);
            }

            return FALSE;
        }
        else if (worst_status == RC_VERIFICATION_STATUS_UNDEF) {
            gboolean ask = gui_prefs_get_signature_ask();

            if (ask) {
                char *filename;
                int text_length;
                char *text;
                GtkWidget *dialog;
                GtkRequisition req;
                int res;

                filename = g_strdup_printf(
                    "        %s\n", package->package_filename);
                text_length = gdk_string_measure(
                    GET_WIDGET("caution_info")->style->font, filename);

                text = g_strdup_printf(
                    _("INFO:\n"
                      "Name: %s (%s)\n"
                      "Version: %s\n"
                      "Release: %s\n"
                      "Package Filename:\n"
                      "%s"),
                    rc_gui_get_pretty_name(package->spec.name), 
                    package->spec.name,
                    package->spec.version, 
                    package->spec.release,
                    filename);
                g_free(filename);
                gtk_label_set_text(
                    GTK_LABEL(GET_WIDGET("caution_info")), text);
                g_free(text);

                dialog = GET_WIDGET("signature_dialog");

                gtk_widget_size_request(dialog, &req);

                if (req.width < text_length)
                    gtk_widget_set_usize(dialog, text_length, -1);

                gnome_dialog_close_hides(GNOME_DIALOG(dialog), TRUE);
                res = gnome_dialog_run_and_close(GNOME_DIALOG(dialog));

                if (res) {
                    text = g_strdup_printf(
                        _("Verification of %s (%s) is inconclusive; "
                          "aborting"),
                        rc_gui_get_pretty_name(package->spec.name),
                        package->spec.name);
                }
                else {
                    text = g_strdup_printf(
                        _("Verification of %s (%s) is inconclusive; "
                          "continuing"),
                        rc_gui_get_pretty_name(package->spec.name),
                        package->spec.name);
                }
                gui_log_append(appinfo->log[LOG_VERIFY], text);
                g_free(text);

                if (res != 0) /* OK Button */
                    return FALSE;
            }
            else {
                char *text;

                text = g_strdup_printf(
                    _("Verification of %s (%s) is inconclusive; continuing"),
                    rc_gui_get_pretty_name(package->spec.name),
                    package->spec.name);
                gui_log_append(appinfo->log[LOG_VERIFY], text);
                g_free(text);
            }
        }
        else { /* Passed */
            char *text;

            if (gpg) {
                text = g_strdup_printf(
                    _("%s (%s) has been verified as cryptographically secure "
                    "by %s"),
                    rc_gui_get_pretty_name(package->spec.name), 
                    package->spec.name, gpg_info);
            }
            else {
                text = g_strdup_printf(
                    _("The integrity of %s (%s) has been confirmed."),
                    rc_gui_get_pretty_name(package->spec.name), 
                    package->spec.name);
            }
            gui_log_append(appinfo->log[LOG_VERIFY], text);
            g_free(text);
        }

        current_count++;
        i = i->next;
    }

    rc_gui_set_progress(PROGRESS_VERIFY, total_count, total_count, NULL);

    return TRUE;
} /* verify_packages */    

static gboolean
transact_packages(RCPackageSList *add_packages,
                  RCPackageSList *remove_packages)
{
    GtkWidget *single, *total;
    int package_count;
    RCPackmanError err;

    gtk_widget_hide(GET_WIDGET("cancel_button"));
    gtk_widget_hide(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));

    package_count =
        g_slist_length(add_packages) + g_slist_length(remove_packages);
    
    single = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "single_progress");
    total = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");
   
    gtk_progress_configure(GTK_PROGRESS(single), 0.0, 0.0, 100.0);
    gtk_progress_configure(
        GTK_PROGRESS(total), 0.0, 0.0, (gfloat) package_count);

    gtk_progress_set_format_string(GTK_PROGRESS(single), "%P%%");
    gtk_progress_set_format_string(GTK_PROGRESS(total), _("%p%% (%v of %u)"));

    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transact_start",
        GTK_SIGNAL_FUNC(transact_start_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transact_step",
        GTK_SIGNAL_FUNC(transact_step_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transact_progress",
        GTK_SIGNAL_FUNC(transact_progress_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transact_done",
        GTK_SIGNAL_FUNC(transact_done_cb), NULL);

    /* This is a blocking operation, but runs the GTK main loop */
    rc_packman_transact(appinfo->pman, add_packages, remove_packages);

    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transact_start_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transact_step_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transact_progress_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transact_done_cb, NULL);

    err = rc_packman_get_error(appinfo->pman);
    if (err) {
        char *text;
        gchar *reason;

        reason = rc_gui_filter_packman_error (
            rc_packman_get_reason (appinfo->pman));

        if (err == RC_PACKMAN_ERROR_FATAL) {
            text = g_strdup_printf(
                _("<p>A fatal error has occurred while Red Carpet was "
                  "processing this transaction. This is likely caused by an "
                  "improperly built package.</p>"
                  "<p>If this problem persists, please report it to the "
                  "<a href=\"http://bugzilla.ximian.com\">Ximian "
                  "Bugzilla</a>.</p>"
                  "<p>Information from the packaging system:<br>%s</p>"),
                reason);
            rc_gui_error_box(text, RC_ERROR_FATAL);
            g_free(text);
            g_free(reason);
            gtk_exit(1);
        }
        else {
            text = g_strdup_printf(
                _("<p>An error has occurred while Red Carpet was processing "
                  "this transaction. This may be caused by an improperly "
                  "built or corrupted package. You may want to try the "
                  "transaction again with fewer packages. You can "
                  "<a href=\"updater:send_bug_report?Transaction failed\">"
                  "send a problem report</a> to Ximian for analysis.</p>"
                  "<p>Information from the packaging system:<br>%s</p>"),
                reason);
            rc_gui_error_box(text, RC_ERROR_ABORT);
            g_free(text);
            g_free(reason);
            return FALSE;
        }
    }

    return TRUE;
} /* transact_packages */

static void
rc_gui_channel_update_all(GUIChannel *current)
{
    GSList *i;
	
    i = appinfo->channel_list;
    while (i) {
        GUIChannel *c = i->data;
        
        if (c->subscribed)
            rc_gui_channel_update_package_lists(c);

        if (!current || current == c) {
            RCPackageSList **packages;

            packages = gtk_object_get_data(
                GTK_OBJECT(appinfo->update_html), "packages_to_install");
            if (packages) {
                g_slist_free(*packages);
                *packages = NULL;
            }

            packages = gtk_object_get_data(
                GTK_OBJECT(appinfo->available_html), "packages_to_install");
            if (packages) {
                g_slist_free(*packages);
                *packages = NULL;
            }

            packages = gtk_object_get_data(
                GTK_OBJECT(appinfo->installed_html), "packages_to_remove");
            if (packages) {
                g_slist_free(*packages);
                *packages = NULL;
            }
        }
        
        i = i->next;
    }

#if 0
    rc_gui_channel_select(current);
    if (!current)
        rc_gui_show_executive_summary();
#endif
} /* rc_gui_channel_update_all */

static void
ok_clicked_cb (GtkWidget *button)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "### OK CLICKED\n");
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("ok_button")), ok_clicked_cb, NULL);
    rc_gui_set_menubar_sensitive (TRUE);
    rc_gui_set_page("main_notebook", PAGE_CHANNELS);
} /* ok_clicked_cb */

static gboolean
build_urls(RCPackageSList *packages, GSList **urls_out,
           guint64 *total_size_out)
{
    GSList *urls = NULL;
    guint64 total_size = 0;
    char *cache_dir;
    GSList *iter;

    /* First create the directory. */
    cache_dir = gui_cache_get_cache_directory(appinfo->package_cache);
    
    if (!g_file_exists(cache_dir)) {
        int err;
        gboolean retry = FALSE;

        do {
            err = rc_mkdir(cache_dir, 0755);
            if (err) {
                gchar *error = g_strdup_printf(
                    _("Unable to create directory %s for local package "
                      "storage"), cache_dir);

                retry = rc_gui_error_box(error, RC_ERROR_RETRY);
                g_free(error);
                if (!retry) {
                    g_free(cache_dir);
                    return FALSE;
                }
            }
        } while (err && retry);
    }
    g_free(cache_dir);

    /* Now, build a list of URLs */
    for (iter = packages; iter; iter = iter->next) {
        RCPackage *pkg = iter->data;
        RCPackageUpdate *upd = rc_package_get_latest_update(pkg);
        char *url;

        total_size += upd->package_size;

        url = rc_gui_get_mirrored_url(
            upd->package_url, RCCHANNEL(pkg->channel));

        /* We have to check to make sure this isn't already on the local
         * file system by checking if it's a "file" URI.  Otherwise, the
         * following call will just work because the package cache just
         * cares about the filename, not anything beofre it.
         */

        if (g_strncasecmp(url, "file://", 7) == 0)
            pkg->package_filename = g_strdup(url + 7);
        else {
            pkg->package_filename = gui_cache_get_cache_filename(
                appinfo->package_cache, url);
        }

        urls = g_slist_prepend(urls, url);

        if (upd->signature_url) {
            total_size += upd->package_size;
        
            url = rc_gui_get_mirrored_url(
                upd->signature_url, RCCHANNEL(pkg->channel));

            if (g_strncasecmp(url, "file://", 7) == 0)
                pkg->signature_filename = g_strdup(url + 7);
            else {
                pkg->signature_filename = gui_cache_get_cache_filename(
                    appinfo->package_cache, url);
            }

            urls = g_slist_prepend(urls, url);
        }
    }

    if (rc_debug_get_log_level() >= RC_DEBUG_LEVEL_INFO) {
        for (iter = urls; iter; iter = iter->next) {
            rc_debug (RC_DEBUG_LEVEL_INFO, "URL: %s\n", (char *) iter->data);
        }
    }

    if (urls_out)
        *urls_out = g_slist_reverse(urls);
    else
        g_slist_free(urls);

    if (total_size_out)
        *total_size_out = total_size;

    return TRUE;
} /* build_urls */

void
rc_gui_verify_installed_packages (void)
{
    RCChannelSList *sublist = NULL;
    RCChannelSList *unsublist = NULL;
    RCPackageSList *added_pkgs = NULL;
    RCPackageSList *removed_pkgs = NULL;
    GSList         *urls = NULL;
    guint64        total_size = 0;
    char           *transaction_id = NULL;
    
    /* Split the list of channels into subscribed and unsubscribed */
    rc_gui_split_channel_list (appinfo->channel_list, &sublist, &unsublist);

    if (!(dependency_check(
              NULL, DEPENDENCY_CHECK_VERIFY,
              NULL, sublist, unsublist,
              &added_pkgs, &removed_pkgs))) {
        /* Either a dependency error or the user didn't approve the 
           resolution */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    if (!(build_urls(added_pkgs, &urls, &total_size))) {
        /* An error prevented a list from being made */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    if (appinfo->priority)
        rc_send_transaction(added_pkgs, removed_pkgs, &transaction_id);
    
    if (added_pkgs)
        rc_gui_prepare_install_page(GUI_INSTALL);
    else
        rc_gui_prepare_install_page(GUI_REMOVE);
    rc_gui_set_page ("main_notebook", PAGE_DOWNLOAD);

    rc_gui_log_display_reset_all();

    if (added_pkgs) {
        /* First do the download */
        if (!download_files(urls, total_size)) {
            if (appinfo->priority)
                rc_send_success(transaction_id, FALSE, "Download failed");
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            rc_gui_set_menubar_sensitive(TRUE);
            goto install_cleanup;
        }

        /* Then verify the packages */
        if (!verify_packages(added_pkgs)) {
            if (appinfo->priority)
                rc_send_success(transaction_id, FALSE, "Verification failed");
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            rc_gui_set_menubar_sensitive (TRUE);
            goto install_cleanup;
        }
    }

    /* Install/Remove and configure */
    if (!transact_packages(added_pkgs, removed_pkgs)) {
        if (appinfo->priority)
            rc_send_success(transaction_id, FALSE, "Transaction failed");
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        rc_gui_set_menubar_sensitive (TRUE);
        goto install_cleanup;
    }

    if (appinfo->priority)
        rc_send_success(transaction_id, TRUE, NULL);
    
    rc_gui_set_progress(PROGRESS_WAIT_FOR_IT, 0, 0, NULL);

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("ok_button")), "clicked",
        GTK_SIGNAL_FUNC(ok_clicked_cb), NULL);

    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), TRUE);
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_grab_focus(GET_WIDGET("ok_button"));

    rc_gui_query_system_packages ();

    rc_gui_channel_update_all(NULL);

    rc_gui_update_world ();

    rc_gui_set_progress(PROGRESS_FINISHED, 0, 0, NULL);

    rc_gui_show_executive_summary();

install_cleanup:
    g_slist_foreach (urls, (GFunc) g_free, NULL);
    g_slist_free (urls);

    g_slist_free (added_pkgs);
    g_slist_free (removed_pkgs);

    g_slist_free (sublist);
    g_slist_free (unsublist);

    rc_gui_set_menubar_sensitive (TRUE);

    rc_gui_show_executive_summary();
} /* rc_gui_verify_installed_packages */

gboolean
gui_install_packages(GUIChannel *current_channel, RCPackageSList *install_pkgs)
{
    RCChannelSList *sublist = NULL;
    RCChannelSList *unsublist = NULL;
    RCPackageSList *pkgs = NULL;
    RCPackageSList *added_pkgs = NULL;
    RCPackageSList *removed_pkgs = NULL;
    GSList         *urls = NULL;
    guint64        total_size = 0;
    int            count;
    gboolean       success = TRUE;
    char           *transaction_id = NULL;

    g_return_val_if_fail(install_pkgs, FALSE);

    /* Split the list of channels into subscribed and unsubscribed */
    rc_gui_split_channel_list (appinfo->channel_list, &sublist, &unsublist);
    
    /* Run it through the dependency system */
    if (!(dependency_check(
              install_pkgs, DEPENDENCY_CHECK_INSTALL,
              current_channel, sublist, unsublist,
              &added_pkgs, &removed_pkgs))) {
        /* This is one of the error cases, dependency_helper should have
           already displayed the error to the user, so we're free to quit at
           this point.  Yay. */
        /* Or the user just didn't agree */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return FALSE;
    }

    pkgs = g_slist_concat(added_pkgs, g_slist_copy(install_pkgs));

    if (!(build_urls(pkgs, &urls, &total_size))) {
        /* An error prevented a list from being made */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return FALSE;
    }

    if (appinfo->priority)
        rc_send_transaction(pkgs, removed_pkgs, &transaction_id);

    rc_gui_prepare_install_page(GUI_INSTALL);
    rc_gui_set_page("main_notebook", PAGE_DOWNLOAD);
    rc_gui_log_display_reset_all();

    /* First do the download */
    if (!download_files(urls, total_size)) {
        if (appinfo->priority)
            rc_send_success(transaction_id, FALSE, "Download failed");
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        rc_gui_set_menubar_sensitive (TRUE);
        success = FALSE;
        goto install_cleanup;
    }

    /* Then verify the packages */
    if (!verify_packages(pkgs)) {
        if (appinfo->priority)
            rc_send_success(transaction_id, FALSE, "Verification failed");
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        rc_gui_set_menubar_sensitive (TRUE);
        success = FALSE;
        goto install_cleanup;
    }

    /* Install and configure */
    if (!transact_packages(pkgs, removed_pkgs)) {
        if (appinfo->priority)
            rc_send_success(transaction_id, FALSE, "Transaction failed");
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        rc_gui_set_menubar_sensitive (TRUE);
        success = FALSE;
        goto install_cleanup;
    }

    if (appinfo->priority)
        rc_send_success(transaction_id, TRUE, NULL);

    count = g_slist_length(pkgs);
    rc_gui_set_progress(PROGRESS_WAIT_FOR_IT, count, count, NULL);

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("ok_button")), "clicked",
        GTK_SIGNAL_FUNC(ok_clicked_cb), NULL);

    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), TRUE);
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_grab_focus(GET_WIDGET("ok_button"));

    rc_gui_query_system_packages ();

    rc_gui_channel_update_all(current_channel);

    rc_gui_update_world ();

    rc_gui_set_progress(PROGRESS_FINISHED, 0, 0, NULL);

install_cleanup:
    g_slist_foreach (urls, (GFunc) g_free, NULL);
    g_slist_free (urls);

    g_slist_free (pkgs);
    g_slist_free (removed_pkgs);

    g_slist_free (sublist);
    g_slist_free (unsublist);

    rc_gui_set_menubar_sensitive (TRUE);

    return success;
} /* gui_install_packages */

void
gui_remove_packages (GUIChannel *channel, RCPackageSList *pkgs)
{
    RCChannelSList *sublist = NULL;
    RCChannelSList *unsublist = NULL;
    RCPackageSList *removed_pkgs = NULL;
    gboolean unsub;
    char *transaction_id = NULL;

    rc_gui_split_channel_list(appinfo->channel_list, &sublist, &unsublist);

    if (!(dependency_check(
              pkgs, DEPENDENCY_CHECK_REMOVE, 
              channel, sublist, unsublist,
              NULL, &removed_pkgs))) {
        /* Either an error occurred or the user didn't confirm the deps 
           check */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    rc_gui_log_display_reset_all();

    unsub = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("unsubscribe1"));
    rc_gui_set_menubar_sensitive (FALSE);

    removed_pkgs = g_slist_concat(removed_pkgs, g_slist_copy(pkgs));

    if (appinfo->priority)
        rc_send_transaction(NULL, removed_pkgs, &transaction_id);

    /* Show the OK button but desensitize it */
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), FALSE);

    rc_gui_prepare_install_page(GUI_REMOVE);
    rc_gui_set_page("main_notebook", PAGE_DOWNLOAD);

    /* Remove the packages */
    if (!transact_packages(NULL, removed_pkgs)) {
        if (appinfo->priority)
            rc_send_success(transaction_id, FALSE, "Transaction failed");
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        goto remove_cleanup;
    }

    if (appinfo->priority)
        rc_send_success(transaction_id, TRUE, NULL);

    rc_gui_set_progress(PROGRESS_WAIT_FOR_IT, 0, 0, NULL);

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("ok_button")), "clicked",
        GTK_SIGNAL_FUNC(ok_clicked_cb), NULL);

    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), TRUE);
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_grab_focus(GET_WIDGET("ok_button"));
    
    rc_gui_query_system_packages ();

    rc_gui_channel_update_all(channel);

    rc_gui_update_world ();

    rc_gui_set_progress(PROGRESS_FINISHED, 0, 0, NULL);

remove_cleanup:
    g_slist_free(removed_pkgs);

    g_slist_free(sublist);
    g_slist_free(unsublist);

    rc_gui_set_menubar_sensitive (TRUE);
    if (unsub)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);
} /* gui_remove_packages */
