/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-uri.c: URI Handler
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <fcntl.h>
#include <crypt.h>
#include <time.h>
#include <pwd.h>

#include <gnome.h>
#include <glade/glade.h>
#include <gtkhtml/gtkhtml.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <libgnomeui/gnome-window-icon.h>

#include <gtkhtml/gtkhtml-stream.h>
#include <gtkhtml/htmlengine-search.h>
#include <gtkhtml/htmlimage.h>
#include <gtkhtml/htmlform.h>

#include <libsoup/soup-uri.h>

#include "gui.h"
#include "gui-log-display.h"
#include "gui-transfer.h"
#include "gui-prefs.h"
#include "gui-install.h"
#include "mirror.h"
#include "subscription.h"

#include "libredcarpet.h"

#include "gui-uri.h"
#include "gui-util.h"
#include "gui-html.h"
#include "gui-glade.h"
#include "gui-channel.h"
#include "gui-priority.h"
#include "gui-debug-report.h"

#include "subscription.h"

static void
action_button_sensitive (gboolean sensitive)
{
    GtkWidget *action_button;

    action_button = (GtkWidget *) gtk_object_get_data (GTK_OBJECT(appinfo->action_html),
                                                       "action_button");

    gtk_widget_set_sensitive (action_button, sensitive ? TRUE : FALSE);
}

void
rc_url_show(const char *url)
{
    char *display, *xauth;
    pid_t pid;
    char *s;

    display = getenv("DISPLAY");
    if (!display) {
        g_warning("Could not get the DISPLAY environment variable.");
        return;
    }

    xauth = rc_gui_get_xauthority_file();
    if (!xauth)
        xauth = getenv("OLD_XAUTHORITY");
    if (!xauth)
        xauth = getenv("XAUTHORITY");

    pid = fork();
    switch (pid) {
    case -1:
        g_warning("Forking of the browser failed");
        break;
    case 0: {
        char *user;
        struct passwd *passwd;
        
        /* Get the USER environment variable */
        user = getenv("USER");
        if (!user) {
            g_warning("Could not get the USER environment variable.");
            _exit(1);
        }

        /* Get the /etc/passwd entry for $USER */
        passwd = getpwnam(user);
        if (!passwd) {
            g_warning("Could not get the /etc/passwd entry for %s", user);
            _exit(1);
        }

        /* Set the user ID for this process so we're not root (unless the
           user actually ran RC as root */
        setuid(passwd->pw_uid);
        
        /* Need to reset $HOME */
	
        s = g_strdup_printf("HOME=%s", passwd->pw_dir);
        putenv(s);
        /* Set DISPLAY */
        s = g_strdup_printf("DISPLAY=%s", display);
        putenv(s);

        /* Set XAUTHORITY, if we have it */
        if (xauth) {
            s = g_strdup_printf("XAUTHORITY=%s", xauth);
            putenv(s);
        }

        gnome_url_show(url);
        _exit(0);
    }
    default:
        wait(NULL);

    }
} /* rc_url_show */

/* CB */
void
rc_gui_html_link_clicked(GtkHTML *html, const char *uri, GUIChannel *channel)
{
    /* This is necessary to keep the keyword search box from constantly
       redrawing at the bottom. */
    gtk_signal_emit_stop_by_name(GTK_OBJECT(html), "link_clicked");

    if (g_strncasecmp(uri, "updater", 7) == 0)
        rc_gui_handle_uri(html, uri, channel);
    else
        rc_url_show(uri);
} /* rc_gui_html_link_clicked */

/* CB */
void
rc_gui_html_load_image(GtkHTML *html, char *url, GtkHTMLStream *stream)
{
    char *das_url = NULL;
    char *base_url;

    g_return_if_fail(html);

    if (g_strncasecmp(url, "updater", 6) == 0) {
        GUIChannel *c;

        if (strncmp (url + strlen ("updater:"), "check", 5) == 0) {
            /* it's a check on/off */
            if (strcmp (url + strlen ("updater:check_"), "inst") == 0) {
                das_url = g_strdup (CKINSTALL_IMAGE_URL);
            } else if (strcmp (url + strlen ("updater:check_"), "rem") == 0) {
                das_url = g_strdup (CKREMOVE_IMAGE_URL);
            } else {
                das_url = g_strdup (CKNOTHING_IMAGE_URL);
            }
        } else {
            c = gtk_object_get_data(GTK_OBJECT(html), "channel");

            gtk_object_set_data(GTK_OBJECT(html), "stream", stream);
            rc_gui_handle_uri(html, url, c);
            return;
        }
    }

    if (!das_url) {
        if (!rc_url_is_absolute(url)) {
            char *p;

            base_url = gtk_object_get_data (GTK_OBJECT (html),
                                            "base_url");
            /* We're not doing mirroring on these URLs. I think that's okay, as
               I think this will only happen in the info pages. Not sure. */


            /* Handle relative URLs for imagitas */
            if (url[0] == '/') {
                SoupUri *purl = soup_uri_new(base_url);

                if (!purl) {
                    rc_debug(RC_DEBUG_LEVEL_WARNING, 
                             __FUNCTION__ "Invalid base_url: %s\n",
                             base_url);
                    return;
                }

                /* Replace the whole path with what we were given */
                g_free(purl->path);
                purl->path = g_strdup(url);
                das_url = soup_uri_to_string(purl, TRUE);
                soup_uri_free(purl);
            }
            else {
                if (base_url) {
                    p = strrchr (base_url, '/');
                    if (p && *(p+1)) {
                        /* URL doesn't end in /, we need to strip the last element */
                        gchar *tmpbase;
                        tmpbase = g_strndup (base_url, p - base_url);
                        das_url = g_strdup_printf ("%s/%s", tmpbase, url);
                        g_free (tmpbase);
                    }
                }
                
                if (!das_url) {
                    das_url = (base_url) 
                        ? g_strdup_printf ("%s/%s", base_url, url)
                        : g_strdup (url);
                }
            }
        }
        else {
                das_url = g_strdup (url);
        }
    }

    rc_debug (RC_DEBUG_LEVEL_INFO, "html_load_image: grabbing %s\n", das_url);

    rc_gui_html_download (html, stream, das_url);

    g_free (das_url);
} /* rc_gui_html_load_image */

/* Note! html might be NULL in some cases (i.e. the buttons at the bottom of each page clicked */
void
rc_gui_handle_uri (GtkHTML *html, const char *uri, GUIChannel *channel)
{
    char **compound;
    char *token;

    g_return_if_fail(uri);

    rc_debug (RC_DEBUG_LEVEL_INFO, "Updater URI clicked: %s\n", uri);

    if (!channel)
        channel = appinfo->current_channel;

    compound = g_strsplit(uri, "#", 1);

    token = strchr(compound[0], ':');
    if (!token) {
        rc_debug (RC_DEBUG_LEVEL_WARNING, "Invalid URI\n");
        g_strfreev(compound);
        return;
    }
    token++;

    /* Note that these have to be in lowercase; we can't downcase, since there
     * might be bits after the main URI that we want to preserve case in.
     */

    if (strcmp(token, "update_page") == 0) {
        rc_gui_set_page("notebook", PAGE_UPDATES);
    }
    else if (strcmp(token, "installed_page") == 0) {
        rc_gui_render_installed_html(channel);
        rc_gui_set_page("notebook", PAGE_INSTALLED);
    }
    else if (strcmp(token, "available_page") == 0) {
        rc_gui_render_available_html(channel);
        rc_gui_set_page("notebook", PAGE_AVAILABLE);
    }
    else if (strcmp(token, "preferences_page") == 0) {
        rc_gui_set_page("notebook", PAGE_PREFERENCES);
    }
    else if (strcmp(token, "info_page") == 0) {
        rc_gui_render_info_html(channel);
        rc_gui_set_page("notebook", PAGE_INFO);
    }
    else if (strcmp(token, "maybe_subscribe") == 0) {
        g_return_if_fail(html);
        g_return_if_fail(channel);

        if (html->engine->form)
            html_form_submit(html->engine->form);
        else
            rc_gui_channel_subscribe(channel);
    }
    else if (strcmp(token, "subscribe") == 0) {
        g_return_if_fail(channel);

        rc_gui_channel_subscribe(channel);
    }
    else if (strcmp(token, "unsubscribe") == 0) {
        g_return_if_fail(channel);

        rc_gui_channel_unsubscribe(channel);
    }
    else if (strcmp(token, "summary") == 0) {
        rc_gui_channel_select(NULL);
        rc_gui_show_executive_summary();
    }
    else if (strcmp(token, "news_page") == 0) {
        rc_gui_channel_select(NULL);
        rc_gui_show_news_page();
    }
    else if (strcmp(token, "featured_page") == 0) {
        rc_gui_channel_select(NULL);
        rc_gui_show_featured_page();
    }
    else if (strcmp(token, "subscribed_page") == 0) {
        rc_gui_channel_select(NULL);
        rc_gui_show_channel_page (TRUE);
    }
    else if (strcmp(token, "unsubscribed_page") == 0) {
        rc_gui_channel_select(NULL);
        rc_gui_show_channel_page (FALSE);
    }
    else if (strncmp(token, "channel_name", 12) == 0) {
        char *s;
        GUIChannel *gc;

        s = strchr(token, '?');
        if (!(s && *++s))
            goto done;

        gc = GUICHANNEL(rc_channel_get_by_name(appinfo->channel_list, s));
        if (!gc) {
            rc_debug (RC_DEBUG_LEVEL_ERROR, "gc == NULL (can't find channel %s)\n", s);
            return;
        }

        rc_gui_channel_select(gc);

        channel = gc;
    }
    else if (strncmp(token, "channel", 7) == 0) {
        char *s;
        int cid;
        GUIChannel *gc;

        s = strchr (token, '?');
        if (!(s && *++s)) 
            goto done;

        cid = atoi(s);
        gc = GUICHANNEL(rc_channel_get_by_id(appinfo->channel_list, cid));
        g_return_if_fail(gc);

        rc_gui_channel_select(gc);

        channel = gc;
    }
    else if (strcmp(token, "update") == 0) {
        RCPackageSList **packages = gtk_object_get_data(
            GTK_OBJECT(appinfo->update_html), "packages_to_install");

        g_return_if_fail(channel);

        if (!packages || !*packages)
            goto done;

        gui_install_packages(channel, *packages);
        rc_gui_channel_select(channel);
    }
    else if (strcmp(token, "install") == 0) {
        RCPackageSList **packages = gtk_object_get_data(
            GTK_OBJECT(appinfo->available_html), "packages_to_install");
        
        g_return_if_fail(channel);

        if (!packages || !*packages)
            goto done;

        gui_install_packages(channel, *packages);
        rc_gui_channel_select(channel);
    }
    else if (strcmp(token, "remove") == 0) {
        RCPackageSList **packages = gtk_object_get_data(
            GTK_OBJECT(appinfo->installed_html), "packages_to_remove");

        if (!packages || !*packages)
            goto done;

        gui_remove_packages(channel, *packages);
        rc_gui_channel_select(channel);
    }
    else if (strcmp(token, "install_all") == 0) {
        if (appinfo->necessary) {
            /* There are necessary updates */
            if (gui_install_packages (NULL, appinfo->necessary)) {
                rc_gui_restart (TRUE);
            }
        }
        else {
            RCPackageSList *uberlist = NULL;
            GSList *i;

            i = appinfo->channel_list;
            while (i) {
                GUIChannel *c = i->data;
                
                if (c->subscribed) {
                    uberlist = g_slist_concat(
                        uberlist, g_slist_copy (c->available_updates));
                }

                i = i->next;
            }
            gui_install_packages(NULL, uberlist);
            rc_gui_show_executive_summary();
            g_slist_free(uberlist);
        }
    }
    else if (strcmp(token, "prefs_proxy") == 0) {
        GtkWidget *dialog;
        GtkWidget *n;

        dialog = GET_WIDGET("property_dialog");
        n = GNOME_PROPERTY_BOX(dialog)->notebook;

        gui_prefs_dialog_show();
        gtk_notebook_set_page(GTK_NOTEBOOK(n), 1);
    }
    else if (strcmp(token, "prefs_cache") == 0) {
        GtkWidget *dialog;
        GtkWidget *n;

        dialog = GET_WIDGET("property_dialog");
        n = GNOME_PROPERTY_BOX(dialog)->notebook;

        gui_prefs_dialog_show();
        gtk_notebook_set_page(GTK_NOTEBOOK(n), 2);
    }
    else if (strcmp(token, "cache_clear") == 0) {
        GtkWidget *w;

        gui_cache_invalidate_all(appinfo->package_cache);

        w = gnome_app_message(
            GNOME_APP(GET_WIDGET("main_window")),
            "The cache has been cleared.");
        gtk_window_set_modal(GTK_WINDOW(w), FALSE);
        gtk_widget_show(w);
    }
    else if (strncmp(token, "info", 4) == 0) {
        char *s;
        RCPackage *package;

        s = strchr(token, '?');
        if (!(s && *++s)) goto done;
        package = (RCPackage *) strtoul(s, NULL, 16);
        rc_gui_render_package_info_html(package);
    }
    else if (strcmp(token, "close") == 0) {
        gtk_widget_hide(GET_WIDGET("package_vbox"));
    }
    else if (strncmp(token, "send_bug_report", 15) == 0) {
        char *s;

        s = strchr(token, '?');
        if (s) /* Move past the '?' */
            s++;

        gui_debug_report_main(s);
    }
    else if (strncmp(token, "install_toggle", 14) == 0) {
        char *s, *p;
        char *imgname;
        RCPackage *pkg;
        RCPackageSList **packages;
        HTMLImage *hi;
        unsigned long a;

        s = strchr (token, '?');
        if (!(s && *++s)) 
            goto done;

        a = strtoul (s, &p, 0);
        imgname = g_strdup_printf ("ckimg%ld", a);

        if (!(p && *++p)) {
            g_free (imgname);
            goto done;
        }

        /* *** FIXME *** this is not 64-bit safe */
        /* this is a pointer address, so... :-P */
        pkg = (RCPackage *) strtoul (p, NULL, 16);
        packages = gtk_object_get_data(
            GTK_OBJECT(html), "packages_to_install");
        hi = gtk_html_get_object_by_id (html, imgname);
        if (!hi) {
            rc_debug(
                RC_DEBUG_LEVEL_WARNING, 
                "Failed in getting htmlimage for %s\n", imgname);
        }

        if (g_slist_find (*packages, pkg)) {
            *packages = g_slist_remove (*packages, pkg);
            if (hi) 
                html_image_set_url(hi, "updater:check_nothing");
        } 
        else {
            *packages = g_slist_append (*packages, pkg);
            if (hi) 
                html_image_set_url(hi, "updater:check_inst");
        }

        action_button_sensitive (g_slist_length (*packages));

        gtk_widget_queue_draw (GTK_WIDGET (html));
        g_free (imgname);
    }
    else if (strncmp(token, "remove_toggle", 13) == 0) {
        char *s, *p;
        char *imgname;
        RCPackage *pkg;
        RCPackageSList **packages;
        HTMLImage *hi;
        unsigned long a;

        s = strchr (token, '?');
        if (!(s && *++s)) 
            goto done;

        a = strtoul (s, &p, 0);
        imgname = g_strdup_printf ("ckimg%ld", a);

        if (!(p && *++p)) {
            g_free (imgname);
            goto done;
        }

        /* *** FIXME *** this is not 64-bit safe */
        /* this is a pointer address, so... :-P */
        pkg = (RCPackage *) strtoul (p, NULL, 16);

        packages = gtk_object_get_data (GTK_OBJECT(html), "packages_to_remove");
        hi = gtk_html_get_object_by_id (html, imgname);
        if (!hi) {
            rc_debug (RC_DEBUG_LEVEL_WARNING, "Failed in getting htmlimage for %s\n", imgname);
        }
        if (g_slist_find (*packages, pkg)) {
            *packages = g_slist_remove (*packages, pkg);
            if (hi) 
                html_image_set_url (hi, "updater:check_nothing");
        } 
        else {
            *packages = g_slist_append (*packages, pkg);
            if (hi) 
                html_image_set_url (hi, "updater:check_rem");
        }

        action_button_sensitive (g_slist_length (*packages));

        gtk_widget_queue_draw (GTK_WIDGET (html));
        g_free (imgname);
    }
    else if (strncmp(token, "install_set", strlen("install_set")) == 0) {
#ifdef GUI_USE_PACKAGESETS
        gchar *sname;
        sname = strchr (token, '?');
        if (p) {
            GSList *l;
            RCPackageSet *ps;
            RCPackageSList **plist;
            int i;

            sname++; /* skip '?' */

            /* Find the package set */
            l = channel->rcchannel.package_sets;
            while (l) {
                ps = (RCPackageSet *) l->data;
                if (strcmp (sname, ps->name) == 0) {
                    /* This one */
                    break;
                }
                l = l->next;
            }

            if (!l) {
                rc_debug (RC_DEBUG_LEVEL_WARNING,
                    "got install_set for non-valid package set: '%s'\n", sname);
                goto done;
            }

            plist = gtk_object_get_data(
                GTK_OBJECT(appinfo->available_html), "packages_to_install");
            if (plist && *plist) {
                g_slist_free(*plist);
                *plist = NULL;
            } else if (!plist) {
                plist = g_new0(RCPackageSList *, 1);
                *plist = NULL;
                gtk_object_set_data(
                    GTK_OBJECT(appinfo->available_html), "packages_to_install",
                    plist);
            }

            /* ps now has this package set, we need to add these packages
             * and simulate a install */
            l = ps->packages;
            while (l) {
                RCPackage *pkg;
                gchar *pname = (gchar *) l->data;
                /* We need to find this package */
                pkg = pkginfo_find_package(&channel->rcchannel, pname);
                if (!pkg) {
                    rc_debug (RC_DEBUG_LEVEL_WARNING, 
                        "Can't find package named '%s' in set '%s' in channel "
                        "id %d\n", pname, ps->name, channel->rcchannel.id);
                    l = l->next;
                    continue;
                }

                *plist = g_slist_append (*plist, pkg);
                l = l->next;
            }

            /* pretend we just hit install */
            rc_gui_handle_uri (html, "updater:install", channel);
        }
#endif
    }
    else if (strcmp(token, "select_all") == 0)
        select_all_or_none(appinfo->current_channel, PAGE_UPDATES, TRUE);
    else if (strcmp(token, "unselect_all") == 0)
        select_all_or_none(appinfo->current_channel, PAGE_UPDATES, FALSE);
    else if (strcmp(token, "select_all_available") == 0)
        select_all_or_none(appinfo->current_channel, PAGE_AVAILABLE, TRUE);
    else if (strcmp(token, "unselect_all_available") == 0)
        select_all_or_none(appinfo->current_channel, PAGE_AVAILABLE, FALSE);
    else if (strcmp(token, "unselect_all_installed") == 0)
        select_all_or_none(appinfo->current_channel, PAGE_INSTALLED, FALSE);
    else if (strcmp(token, "download_log") == 0)
        rc_gui_log_display(LOG_DOWNLOAD);
    else if (strcmp(token, "verify_log") == 0)
        rc_gui_log_display(LOG_VERIFY);
    else if (strcmp(token, "transact_log") == 0)
        rc_gui_log_display(LOG_TRANSACT);
    else if (strcmp(token, "remove_log") == 0)
        rc_gui_log_display(LOG_REMOVE);
    else if (strncmp(token, "search", 6) == 0) {
        char *s;

        s = strchr (token, '?');
        if (!(s && *++s)) 
            goto done;
        rc_gui_do_search(s);
    }
    else if (strcmp(token, "priority") == 0) {
        rc_gui_enter_activation_code();
    }
    else if (strcmp(token, "disable_priority") == 0) {
        gui_prefs_set_priority(FALSE);
        gui_prefs_sync_priority();
        rc_gui_priority();
        rc_gui_refresh_all_channels();
    }
    else if (strncmp(token, "web", 3) == 0) {
        char *s;
        char *h;
        GtkHTMLStream *stream;

        s = strchr(token, '?');
        if (!(s && *++s))
            goto done;

        stream = gtk_html_begin(html);
        h = g_strdup_printf(
            "<html><body><b>%s</b></body></html>",
            _("Please wait, loading page..."));
        gtk_html_write(html, stream, h, strlen(h));
        gtk_html_end(html, stream, GTK_HTML_STREAM_OK);
        g_free(h);
        rc_gui_html_download(html, NULL, s);

        /* FIXME: Valid HTTP URLs can contain octothorpes, as they indicate
           anchors. Instead of doing the right thing and escaping them, we're
           just going to prohibit compound URIs beyond the web URI. This'll
           get fixed if there is ever a compelling reason for them. */
        g_strfreev(compound);
        return;
    }
    else if (strcmp(token, "restart") == 0) {
        rc_gui_restart(FALSE);
    }
    else if (strcmp(token, "noop") == 0) {
        /* No-op */
    }
    else
        rc_debug(RC_DEBUG_LEVEL_INFO, "Unknown updater URI: `%s'\n", token);

  done:
    /* If there is a compound URI, we want to recurse. */
    if (compound[1] && html)
        rc_gui_handle_uri(html, compound[1], channel);
        
    g_strfreev(compound);
} /* rc_gui_handle_uri */

void
select_all_or_none(
    GUIChannel *channel, MinorNotebookPage current_page , gboolean all)
{
    int count, count_max;
    char *key;
    HTMLImage *hi;
    RCPackageSList **packages;
    GtkHTML *html_type = NULL;
    RCPackageSList *channel_available = NULL;

    switch (current_page) {
    case PAGE_UPDATES:
        html_type = appinfo->update_html;
        channel_available = channel->available_updates;
        break;
    case PAGE_AVAILABLE:
        html_type = appinfo->available_html;
        channel_available = channel->uninstalled_packages;
        break;
    case PAGE_INSTALLED:
        html_type = appinfo->installed_html;
        channel_available = channel->available_updates;
        break;
    default:
        g_assert_not_reached();
        break;
    }
      
    count_max = GPOINTER_TO_INT(
        gtk_object_get_data(GTK_OBJECT(html_type), "checkbox_count"));
    key = g_malloc(100);       /* the biggest int will fit in this */
    for (count = 0; count < count_max; count++) {
        sprintf(key, "ckimg%d", count);
        hi = gtk_html_get_object_by_id(html_type, key);
        html_image_set_url(
            hi, all ? "updater:check_inst" :  "updater:check_nothing");
    }
    g_free (key);

    if (current_page == PAGE_INSTALLED) {
      packages = gtk_object_get_data(
          GTK_OBJECT (html_type), "packages_to_remove");
    }
    else {
      packages = gtk_object_get_data(
          GTK_OBJECT (html_type), "packages_to_install");
    }
    g_slist_free (*packages);

    if (all)
        *packages = g_slist_copy(channel_available);
    else 
        *packages = NULL;

    action_button_sensitive (g_slist_length (*packages));

    /* And then we render */
    gtk_widget_queue_draw (GTK_WIDGET (html_type));
    
} /* select_all_general_none */



typedef struct {
    GtkHTML *html;
    GtkHTMLStream *stream;
    gboolean close;
} GUIHTMLandStream;


/* CB */
static void
render_html_cb(GUITransfer *t, char *buf, int size,
               GUIHTMLandStream *ghas)
{
    if (!ghas->stream)
        ghas->stream = gtk_html_begin(ghas->html);

    if (size > 0)
        gtk_html_write(ghas->html, ghas->stream, buf, size);
} /* render_html_cb */

static gboolean
destroy_async_transfer(gpointer data)
{
    GUITransfer *t = GUI_TRANSFER(data);

    gtk_object_unref(GTK_OBJECT(t));

    return FALSE;
} /* destroy_async_transfer */

/* CB */
static void
finish_html_cb(GUITransfer *t, GUIHTMLandStream *ghas)
{
    GUITransferError err = gui_transfer_get_error(t);

    if (err != GUI_TRANSFER_ERROR_NONE) {
        char *blah = g_strdup_printf(
            "<h2>%s</h2>"
            "URL: %s",
            gui_transfer_get_error_string(t), t->url);
        GtkHTMLStream *stream;

        if (!ghas->stream)
            stream = gtk_html_begin(ghas->html);
        else
            stream = ghas->stream;

        gtk_html_write(ghas->html, stream, blah, strlen(blah));
        gtk_html_end(ghas->html, stream, GTK_HTML_STREAM_ERROR);

        g_free(blah);
    }
    else {
        /* Ew ew ew ew ew ew ewwwwwwwwww! */
        HTMLImagePointer *ip;
        GdkPixbufLoader *loader;

        if (ghas->stream) {
            ip = ghas->stream->user_data;
            loader = ip->loader;
            
            if (!ghas->close) {
                gtk_object_ref(GTK_OBJECT(loader));
                ip->refcount++;
            }
            
            gtk_html_end(ghas->html, ghas->stream, GTK_HTML_STREAM_OK);
            
            if (!ghas->close) {
                ip->loader = loader;
            }
        }
    }

    g_free (ghas);
    gtk_idle_add(destroy_async_transfer, t);
} /* finish_html_cb */

void
rc_gui_html_download(GtkHTML *html, GtkHTMLStream *stream_in, gchar *url)
{
    char *u;

    g_return_if_fail(html);

    /* This is not an error. */
    if (!url)
        return;

    if (!rc_url_is_absolute(url) &&
        gtk_object_get_data(GTK_OBJECT(html), "base_url")) {
        u = g_strconcat(
            gtk_object_get_data(GTK_OBJECT(html), "base_url"), url, NULL);
    }
    else
        u = g_strdup(url);

    if (rc_url_is_absolute(u)) {
        GUITransfer *t = gui_transfer_new(
            GUI_TRANSFER_FLAGS_NONE, appinfo->normal_cache);
        GUIHTMLandStream *ghas = g_new0 (GUIHTMLandStream, 1);

        ghas->stream = stream_in;
        ghas->html = html;
        ghas->close = TRUE;

        /* This is an evil hack because of the in-place changing of images
           that we're doing. */
        if (strcmp(url, CKINSTALL_IMAGE_URL) == 0 ||
            strcmp(url, CKREMOVE_IMAGE_URL) == 0 ||
            strcmp(url, CKNOTHING_IMAGE_URL) == 0)
            ghas->close = FALSE;

        gtk_signal_connect(
            GTK_OBJECT(t), "file_data",
            GTK_SIGNAL_FUNC(render_html_cb), ghas);

        gtk_signal_connect(
            GTK_OBJECT(t), "file_done",
            GTK_SIGNAL_FUNC(finish_html_cb), ghas);

        gui_transfer_begin(t, u);
    } 
    else {
        int fd;
        int bytes;
        char buf[1024];
        GtkHTMLStream *stream;

        if (!stream_in)
            stream = gtk_html_begin(html);
        else
            stream = stream_in;

        fd = open(u, O_RDONLY);
        if (fd < 0) {
            rc_debug(RC_DEBUG_LEVEL_WARNING, "Unable to load image `%s'\n", u);
            gtk_html_end(html, stream, GTK_HTML_STREAM_ERROR);
            g_free(u);
            return;
        }

        do {
            bytes = read(fd, buf, 1024);
            if (bytes > 0) {
                gtk_html_write(html, stream, buf, bytes);
            }
        } while (bytes > 0);

        if (bytes < 0) {
            gtk_html_end(html, stream, GTK_HTML_STREAM_ERROR);
        } else {
            gtk_html_end(html, stream, GTK_HTML_STREAM_OK);
        }
    }

    g_free(u);
} /* rc_gui_html_download */

static void
submit_done(GUITransfer *t)
{
    gtk_object_unref(GTK_OBJECT(t));
} /* submit_done */

void
rc_gui_html_submit(GtkHTML *html, const char *form_method, 
                   const char *form_action, const char *form_encoding,
                   gpointer data)
{
    char *method, *action, *encoding;
    GtkHTMLStream *stream;
    char *tmp;
    char *base_url, *path;
    SoupUri *url;
    char *url_string;
    GUITransfer *t;

    /* Clearing out the HTML page below destroys the form and thus the
       method, action, and encoding. Let's copy them before that happens,
       shall we? */
    method = g_strdup(form_method);
    action = g_strdup(form_action);
    encoding = g_strdup(form_encoding);

    /* Clear out the HTML page and tell them to wait... */
    stream = gtk_html_begin(html);
    tmp = g_strdup_printf(
        "<html><body><b>%s</b></body></html>",
        _("Please wait, loading page..."));
    gtk_html_write(html, stream, tmp, strlen(tmp));
    gtk_html_end(html, stream, GTK_HTML_STREAM_OK);

    /* Mmmm... URL mangling... */
    base_url = gtk_object_get_data(GTK_OBJECT(html), "base_url");
    url = soup_uri_new(base_url);
    path = g_strndup(url->path, (strrchr(url->path, '/') + 1) - url->path);
    g_free(url->path);
    url->path = g_strdup_printf("%s%s?%s", path, action, encoding);
    g_free(path);
    url_string = soup_uri_to_string(url, TRUE);
    soup_uri_free(url);

    /* FIXME: We might want to do a POST HTTP request someday. I dunno why,
       but we might. GUITransfer has no support for this. But
       protocol-specific data makes the baby Jesus cry. Anyway, the point of
       this comment is that we're ignoring what method is set to. */
    t = gui_transfer_new(GUI_TRANSFER_FLAGS_NONE, appinfo->normal_cache);
    gtk_signal_connect(
        GTK_OBJECT(t), "file_done",
        GTK_SIGNAL_FUNC(submit_done), NULL);
    gui_transfer_begin(t, url_string);

    g_free(url_string);
    g_free(method);
    g_free(action);
    g_free(encoding);
} /* rc_gui_html_submit */
